/*
 *  Terminal IIIMCF Interface/Terminal IIIMCF Input Method
 *  Copyright (c) 2005 Young Chol Song <nskystars at yahoo.com>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer
 *     in the documentation and/or other materials provided with the
 *     distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 *  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 *  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iiimcf.h>
#include <time.h>

#include "tii.h"

/*
 * internal function definitions
 */

int tii_fwd_event(IIIMCF_context cx, IIIMCF_event ev);
void tii_process_lookup_choices(IIIMCF_lookup_choice lc);
void tii_debug_print_event(IIIMCF_event ev);

/*
 *  The following functions are code based on other programs;
 *  please refer to their respective licenses for copying.
 *
 *  termim 1.0.0
 *     int wchar_to_utf8(wchar_t c, char * outbuf, int bufsize);
 *
 *  iiimf 12.2
 *     char * card16_to_utf8(const IIIMP_card16 * pu16);
 *     char * iiimcf_text_to_utf8(IIIMCF_text t);
 */

/*
 * Terminal Input Method
 * (C) Nicolas George, 2002-11-03
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* edited from termim.c in termim-1.0.0 */
int
wchar_to_utf8(wchar_t c, char * outbuf, int bufsize)
{
	int n, i;

	if(c < 0x80)
		n = 0;
	else if(c < 0x800)
		n = 1;
	else if(c < 0x10000)
		n = 2;
	else if(c < 0x200000)
		n = 3;
	else if(c < 0x4000000)
		n = 4;
	else
		n = 5;

	if((n + 1) > bufsize)
		return 0;

	*(outbuf++) = (c >> (6 * n)) | (n == 0 ? 0 : (0xFC << (5 - n)));
	for(i = n * 6 - 6; i >= 0; i -= 6)
		*(outbuf++) = ((c >> i) & 0x3F) | 0x80;
	*outbuf = 0;

	return (n + 1);
}

/*
 * Copyright 1990-2001 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions: The above copyright notice and this
 * permission notice shall be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE OPEN GROUP OR SUN MICROSYSTEMS, INC. BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
 * THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE EVEN IF
 * ADVISED IN ADVANCE OF THE POSSIBILITY OF SUCH DAMAGES.
 */

#define CONVERT_BUFSIZE 7
/* edited from lib/iiimcf/test/check/common.c in iiimf-12.2 */
char *
card16_to_utf8(const IIIMP_card16 * pu16)
{
	char *p, *pr, *pr2;
	int i, rest;
	int initlen = 256;
	char utf8buf[CONVERT_BUFSIZE] = {0, };

	pr = (char *)malloc(sizeof(*p) * initlen);
	rest = initlen - 1;
	for(p = pr; *pu16; pu16++) {
		if(*pu16 < 128) {
			if(rest < 1) {
				rest += initlen;
				initlen *= 2;
				pr2 = (char *)realloc(pr, sizeof(*p) * initlen);
				p = pr2 + (p - pr);
				pr = pr2;
			}
			*p++ = *pu16;
			rest--;
		}
		else {
			if(rest < 16) {
				rest += initlen;
				initlen *= 2;
				pr2 = (char *)realloc(pr, sizeof(*p) * initlen);
				p = pr2 + (p - pr);
				pr = pr2;
			}
			wchar_to_utf8((wchar_t)*pu16, utf8buf, CONVERT_BUFSIZE);
			i = snprintf(p, CONVERT_BUFSIZE, "%s", utf8buf);
			p += i;
			rest -= i;
		}
	}
	*p = '\0';

	return pr;
}

/* edited from lib/iiimcf/test/check/common.c in iiimf-12.2 */
char *
iiimcf_text_to_utf8(IIIMCF_text t)
{
	IIIMF_status st;
	char *p, *pr, *pr2;
	int i, n, l, rest, nfb, initlen = 256;
	IIIMP_card16 ch;
	const IIIMP_card32 *pids, *pfbs;
	char utf8buf[CONVERT_BUFSIZE] = {0, };
	
	pr = p = (char *)malloc(sizeof(*p) * initlen);
	rest = initlen - 1;

	st = iiimcf_get_text_length(t, &n);
	if(st != IIIMF_STATUS_SUCCESS)
		return 0;

	for(i = 0; i < n; i++) {
		st = iiimcf_get_char_with_feedback(t, i, &ch, &nfb, &pids, &pfbs);
		if(st != IIIMF_STATUS_SUCCESS)
			return 0;

		if(rest <= (16 + 2 + 20 * nfb)) {
			rest += initlen;
			initlen *= 2;
			pr2 = (char *)realloc(pr, sizeof(*p) * initlen);
			p = pr2 + (p - pr);
			pr = pr2;
		}
		if(ch < 128) {
			*p++ = ch;
			rest--;
		}
		else {
			wchar_to_utf8((wchar_t)ch, utf8buf, CONVERT_BUFSIZE);
			l = snprintf(p, CONVERT_BUFSIZE, "%s", utf8buf);
			p += l;
			rest -= l;
		}
	}
	*p = '\0';

	return pr;
}

void
default_error_handler(char * msg, int no)
{
	fprintf(stderr, "[error] %s (%d)\r\n", msg, no);
}

void
default_message_handler(char * msg, char * item)
{
	fprintf(stderr, "[message] %s (%s)\r\n", msg, item);
}

void
default_lookup_handler(char * title_str, int win_choices, int win_rows, int win_cols,
			int win_dir, int ncand, int curr_idx, char ** cand_label, char ** cand_str)
{
	int i = 0;

	fprintf(stderr, "Title string %s\r\n", title_str);
	fprintf(stderr, "Window choices %d, rows %d, cols %d, ", win_choices, win_rows, win_cols);
	if(win_dir == IIIMCF_LOOKUP_CHOICE_HORIZONTAL_DIRECTION)
		fprintf(stderr, "Horizontal direction\r\n");
	else if(win_dir == IIIMCF_LOOKUP_CHOICE_VERTICAL_DIRECTION)
		fprintf(stderr, "Vertical direction\r\n");
	else
		fprintf(stderr, "Unknown direction\r\n");
	fprintf(stderr, "Number of candidates %d, current candidate %d\r\n", ncand, curr_idx);
	for(i = 0; i < ncand; i++)
		fprintf(stderr, "Candidate %d, [%s] [%s]\r\n", i, cand_label[i], cand_str[i]);
}

void
default_preedit_handler(char * str, int cur_pos)
{
	fprintf(stderr, "Preedit [%s](%d)\r\n", str, cur_pos);
}

void
default_commit_handler(char * str)
{
	fprintf(stderr, "Commit [%s]\r\n", str);
}

void
default_status_handler(char * str)
{
	fprintf(stderr, "Status text [%s]\r\n", str);
}

void
default_keyevent_handler(int keycode, char keychar, int mod)
{
	fprintf(stderr, "Keyevent code %d, char %d, mod %d\r\n", keycode, keychar, mod);
}

void (*error_handler)(char *, int) = default_error_handler;
void (*message_handler)(char *, char *) = default_message_handler;
void (*lookup_handler)(char *, int, int, int, int, int, int, char **, char **) = default_lookup_handler;
void (*preedit_handler)(char *, int) = default_preedit_handler;
void (*commit_handler)(char *) = default_commit_handler;
void (*status_handler)(char *) = default_status_handler;
void (*keyevent_handler)(int, char, int) = default_keyevent_handler;

void
tii_register_error_handler(void (*handler_func)(char *, int))
{
	error_handler = handler_func;
}

void
tii_register_message_handler(void (*handler_func)(char *, char *))
{
	message_handler = handler_func;
}

void
tii_register_lookup_handler(void (*handler_func)(char *, int, int, int, int, int, int, char **, char **))
{
	lookup_handler = handler_func;
}

void
tii_register_preedit_handler(void (*handler_func)(char *, int))
{
	preedit_handler = handler_func;
}

void
tii_register_commit_handler(void (*handler_func)(char *))
{
	commit_handler = handler_func;
}

void
tii_register_status_handler(void (*handler_func)(char *))
{
	status_handler = handler_func;
}

void
tii_register_keyevent_handler(void (*handler_func)(int, char, int))
{
	keyevent_handler = handler_func;
}

int
tii_fwd_event(IIIMCF_context cx, IIIMCF_event ev)
{
	IIIMF_status st;
	char * msg;

	st = iiimcf_forward_event(cx, ev);
	if(st == IIIMF_STATUS_EVENT_NOT_FORWARDED) {
#if DEBUG
		msg = strdup("Event not forwarded");
		(*error_handler)(msg, st);
		free(msg);
#endif
		return 0; // need to send key directly
	}
	else if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Failed to forward event");
		(*error_handler)(msg, st);
		free(msg);
		return 0;
	}

	return 1;
}

int
tii_fwd_trigger(IIIMCF_context cx, int flag)
{
	IIIMF_status st;
	char * msg;

	IIIMCF_event ev;

	st = iiimcf_create_trigger_notify_event(flag, &ev);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Failed creating IM trigger");
		(*error_handler)(msg, st);
		free(msg);
		return 0;
	}
	if(!tii_fwd_event(cx, ev))
		return 0;

	return 1;
}

int
tii_fwd_key(IIIMCF_context cx, int keycode, int keychar, int mod)
{
	IIIMF_status st;
	char * msg;

	IIIMCF_event ev;
	IIIMCF_keyevent kev;

	kev.keycode = keycode;
	kev.keychar = keychar;
	kev.modifier = mod;
	kev.time_stamp = time(NULL);

	st = iiimcf_create_keyevent(&kev, &ev);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Failed creating keyevent");
		(*error_handler)(msg, st);
		free(msg);
		return 0;
	}
	if(!tii_fwd_event(cx, ev))
		return 0; // need to send key directly

	return 1;
}

void
tii_process_lookup_choices(IIIMCF_lookup_choice lc)
{
	IIIMF_status st;
	char * msg;

	IIIMCF_text lctitle;
	char * title_str = NULL;

	int nchoice = 0, rows = 0, cols = 0, dir = -1;

	int ncand = 0, first_idx, last_idx, curr_idx, i, flag;
	IIIMCF_text cand, label;
	char ** cand_label, ** cand_str;

	st = iiimcf_get_lookup_choice_title(lc, &lctitle);
	if(st == IIIMF_STATUS_SUCCESS) {
		char * str;

		str = iiimcf_text_to_utf8(lctitle);
		title_str = strdup(str);
		free(str);
	}
	else {
		msg = strdup("Failed retrieving lookup choice title");
		(*error_handler)(msg, st);
		free(msg);
	}

	st = iiimcf_get_lookup_choice_configuration(lc, &nchoice, &rows, &cols, &dir);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Failed retrieving lookup choice configuration");
		(*error_handler)(msg, st);
		free(msg);
	}

	st = iiimcf_get_lookup_choice_size(lc, &ncand, &first_idx, &last_idx, &curr_idx);
	if(st == IIIMF_STATUS_SUCCESS) {
		cand_label = (char **)malloc(sizeof(char *) * (ncand + 1));
		cand_str = (char **)malloc(sizeof(char *) * (ncand + 1));
		for(i = first_idx; i <= last_idx; i++) {
			st = iiimcf_get_lookup_choice_item(lc, i, &cand, &label, &flag);
			if(st == IIIMF_STATUS_SUCCESS) {
				if(flag & IIIMCF_LOOKUP_CHOICE_ITEM_ENABLED) {
					char * str;

					str = iiimcf_text_to_utf8(label);
					cand_label[i] = strdup(str);
					free(str);
					str = iiimcf_text_to_utf8(cand);
					cand_str[i] = strdup(str);
					free(str);
				}
			}
			else {
				msg = strdup("Error retrieving lookup item");
				(*error_handler)(msg, st);
				free(msg);
			}
		}
		cand_label[i] = NULL;
		cand_str[i] = NULL;
	}
	else {
		msg = strdup("Error retrieving lookup size");
		(*error_handler)(msg, st);
		free(msg);

		cand_label = (char **)malloc(sizeof(char *));
		cand_label[0] = NULL;
		cand_str = (char **)malloc(sizeof(char *));
		cand_str[0] = NULL;
	}

	(*lookup_handler)(title_str, nchoice, rows, cols, dir, ncand, curr_idx, cand_label, cand_str);

	free(title_str);
	tii_free_list(cand_label);
	tii_free_list(cand_str);
}


int
tii_process_events(IIIMCF_context cx)
{
	IIIMF_status st;
	char * msg;

	IIIMCF_event ev;
	IIIMCF_event_type et;

	/* IIIMCF_EVENT_TYPE_KEYEVENT */
	IIIMCF_keyevent kev;

	/* IIIMCF_EVENT_TYPE_UI_PREEDIT_CHANGE */
	int cur_pos;

	/* IIIMCF_EVENT_TYPE_UI_PREEDIT_CHANGE
	   IIIMCF_EVENT_TYPE_UI_COMMIT
	   IIIMCF_EVENT_TYPE_UI_STATUS_CHANGE */
	IIIMCF_text text;
	char * str;

	/* IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_CHANGE */
	IIIMCF_lookup_choice lc;

	while((st = iiimcf_get_next_event(cx, &ev)) == IIIMF_STATUS_SUCCESS) {
#if DEBUG
		tii_debug_print_event(ev);
#endif
		st = iiimcf_get_event_type(ev, &et);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't get event type");
			(*error_handler)(msg, st);
			free(msg);
			return 0;
		}

		switch(et) {
		case IIIMCF_EVENT_TYPE_KEYEVENT:
			st = iiimcf_get_keyevent_value(ev, &kev);
			if(st == IIIMF_STATUS_SUCCESS) {
				(*keyevent_handler)(kev.keycode, kev.keychar, kev.modifier);
			}
			else {
				msg = strdup("Unknown error on keyevent");
				(*error_handler)(msg, st);
				free(msg);
			}
			break;
		case IIIMCF_EVENT_TYPE_UI_PREEDIT_CHANGE:
			st = iiimcf_get_preedit_text(cx, &text, &cur_pos);
			if(st == IIIMF_STATUS_SUCCESS) {
				str = iiimcf_text_to_utf8(text);
				(*preedit_handler)(str, cur_pos);
				free(str);
			}
			else if(st == IIIMF_STATUS_NO_PREEDIT) {
				msg = strdup("No preedit");
				(*error_handler)(msg, st);
				free(msg);
			}
			else {
				msg = strdup("Unknown error on preedit");
				(*error_handler)(msg, st);
				free(msg);
			}
			break;
		case IIIMCF_EVENT_TYPE_UI_COMMIT:
			st = iiimcf_get_committed_text(cx, &text);
			if(st == IIIMF_STATUS_SUCCESS) {
				str = iiimcf_text_to_utf8(text);
				(*commit_handler)(str);
				free(str);
			}
			else if(st == IIIMF_STATUS_NO_COMMITTED_TEXT) {
				msg = strdup("No committed text");
				(*error_handler)(msg, st);
				free(msg);
			}
			else {
				msg = strdup("Unknown error on commit");
				(*error_handler)(msg, st);
				free(msg);
			}
			break;
		case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_CHANGE:
			st = iiimcf_get_lookup_choice(cx, &lc);
			if(st == IIIMF_STATUS_SUCCESS) {
				tii_process_lookup_choices(lc);
			}
			else if(st == IIIMF_STATUS_NO_LOOKUP_CHOICE) {
				msg = strdup("No lookup choice");
				(*error_handler)(msg, st);
				free(msg);
			}
			else {
				msg = strdup("Unknown error on lookup choice");
				(*error_handler)(msg, st);
				free(msg);
			}
			break;
		case IIIMCF_EVENT_TYPE_UI_STATUS_CHANGE:
			st = iiimcf_get_status_text(cx, &text);
			if(st == IIIMF_STATUS_SUCCESS) {
				str = iiimcf_text_to_utf8(text);
				(*status_handler)(str);
				free(str);
			}
			else if(st == IIIMF_STATUS_NO_STATUS_TEXT) {
				msg = strdup("No status text");
				(*error_handler)(msg, st);
				free(msg);
			}
			else {
				msg = strdup("Unknown error on status");
				(*error_handler)(msg, st);
				free(msg);
			}
			break;
		default:
			break;
		}
		iiimcf_ignore_event(ev);
	}

	return 1;
}

char **
tii_get_im_list(IIIMCF_handle hx)
{
	IIIMF_status st;
	char * msg;

	int num_im, i;
	IIIMCF_input_method * im;
	const IIIMP_card16 * im_id;

	char ** list;

	st = iiimcf_get_supported_input_methods(hx, &num_im, &im);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't get IM list");
		(*error_handler)(msg, st);
		free(msg);

		list = (char **)malloc(sizeof(char *));
		list[0] = NULL;
		return list;
	}

	list = (char **)malloc(sizeof(char *) * (num_im + 1));

	for(i = 0; i < num_im; i++) {
		st = iiimcf_get_input_method_desc(im[i], &im_id, NULL, NULL);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't get IM description");
			(*error_handler)(msg, st);
			free(msg);

			list[i] = NULL; // fix me, this ends further printing of IM
			continue;
		}
		list[i] = strdup(card16_to_utf8(im_id));
	}
	list[i] = NULL;

	return list;
}

void
tii_free_list(char ** list)
{
	int i;

	for(i = 0; list[i] != NULL; i++)
		free(list[i]);
	free(list[i]);
	free(list);
}

IIIMCF_input_method
tii_find_im(IIIMCF_handle hx, char * im_str)
{
	IIIMF_status st;
	char * msg;

	int num_im, i;
	IIIMCF_input_method * im;
	const IIIMP_card16 * im_id;

	st = iiimcf_get_supported_input_methods(hx, &num_im, &im);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't get IM list");
		(*error_handler)(msg, st);
		free(msg);

		return NULL;
	}

	for(i = 0; i < num_im; i++) {
		st = iiimcf_get_input_method_desc(im[i], &im_id, NULL, NULL);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't get IM list");
			(*error_handler)(msg, st);
			free(msg);

			continue;
		}
		if(!strcmp(card16_to_utf8(im_id), im_str)) {
			msg = strdup("Found IM");
			(*message_handler)(msg, im_str);
			free(msg);
			return im[i];
		}
	}

	msg = strdup("Can't find IM");
	(*message_handler)(msg, im_str);
	free(msg);
	return NULL;
}

char **
tii_get_lang_list(IIIMCF_input_method im)
{
	IIIMF_status st;
	char * msg;

	int num_lang, i;
	IIIMCF_language * lang;

	char ** list;
	const char * lang_id;

	st = iiimcf_get_input_method_languages(im, &num_lang, &lang);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't retrieve IM lang list");
		(*error_handler)(msg, st);
		free(msg);

		list = (char **)malloc(sizeof(char *));
		list[0] = NULL;
		return list;
	}

	list = (char **)malloc(sizeof(char *) * (num_lang + 1));

	for(i = 0; i < num_lang; i++) {
		st = iiimcf_get_language_id(lang[i], &lang_id);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't retrieve IM lang list");
			(*error_handler)(msg, st);
			free(msg);

			list[i] = NULL; // fix me, this stops further printing of lang
			continue;
		}
		list[i] = strdup(lang_id);
	}
	list[i] = NULL;

	return list;
}

IIIMCF_language
tii_find_lang(IIIMCF_handle hx, char * im_str, char * lang_str)
{
	IIIMF_status st;
	char * msg;

	int i;

	IIIMCF_input_method im;

	int num_lang;
	IIIMCF_language * lang;
	const char * lang_id;

	im = tii_find_im(hx, im_str);

	st = iiimcf_get_input_method_languages(im, &num_lang, &lang);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't get IM lang list");
		(*error_handler)(msg, st);
		free(msg);
		return NULL;
	}

	for(i = 0; i < num_lang; i++) {
		st = iiimcf_get_language_id(lang[i], &lang_id);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't get language id");
			(*error_handler)(msg, st);
			free(msg);
			continue;
		}
		if(!strcmp(lang_id, lang_str)) {
			msg = strdup("Found lang");
			(*message_handler)(msg, lang_str);
			free(msg);
			return lang[i];
		}
	}

	msg = strdup("Can't find lang");
	(*message_handler)(msg, lang_str);
	free(msg);
	return NULL;
}

int
tii_get_conversion_mode(IIIMCF_context cx)
{
	IIIMF_status st;
	int flag;

	st = iiimcf_get_current_conversion_mode(cx, &flag);
	if(st != IIIMF_STATUS_SUCCESS)
		return -1;

	return flag;
}

int
tii_init_iiimcf(void)
{
	IIIMF_status st;
	char * msg;

	st = iiimcf_initialize(IIIMCF_ATTR_NULL);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't initialize IIIMCF");
		(*error_handler)(msg, st);
		free(msg);

		return 0;
	}

	return 1;
}

IIIMCF_handle
tii_create_handle(void)
{
	IIIMF_status st;
	char * msg;

	IIIMCF_handle hx;

	st = iiimcf_create_handle(IIIMCF_ATTR_NULL, &hx);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't create handle. Did you execute iiimd (or htt_server)?");
		(*error_handler)(msg, st);
		free(msg);

		iiimcf_finalize();
		return NULL;
	}

	return hx;
}

IIIMCF_context
tii_create_context(IIIMCF_handle hx, IIIMCF_input_method im, IIIMCF_language lang, int trigger_auto)
{
	IIIMF_status st;
	char * msg;

	IIIMCF_attr attr;
	IIIMCF_context cx;

	st = iiimcf_create_attr(&attr);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't create attribute");
		(*error_handler)(msg, st);
		free(msg);

		iiimcf_destroy_handle(hx);
		iiimcf_finalize();
		return NULL;
	}

	if(im != NULL) {
		st = iiimcf_attr_put_ptr_value(attr, IIIMCF_ATTR_INPUT_METHOD, im);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't put pointer value in attribute");
			(*error_handler)(msg, st);
			free(msg);

			iiimcf_destroy_attr(attr);
			iiimcf_destroy_handle(hx);
			iiimcf_finalize();
			return NULL;
		}
	}

	if(lang != NULL) {
		st = iiimcf_attr_put_ptr_value(attr, IIIMCF_ATTR_INPUT_LANGUAGE, lang);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't put pointer value in attribute");
			(*error_handler)(msg, st);
			free(msg);

			iiimcf_destroy_attr(attr);
			iiimcf_destroy_handle(hx);
			iiimcf_finalize();
			return NULL;
		}
	}

	if(trigger_auto == 0 || trigger_auto == 1) {
		st = iiimcf_attr_put_integer_value(attr, IIIMCF_ATTR_DISABLE_AUTOMATIC_TRIGGER_NOTIFY, trigger_auto);
		if(st != IIIMF_STATUS_SUCCESS) {
			msg = strdup("Can't put integer value in attribute");
			(*error_handler)(msg, st);
			free(msg);

			iiimcf_destroy_attr(attr);
			iiimcf_destroy_handle(hx);
			iiimcf_finalize();
			return NULL;
		}
	}

	st = iiimcf_create_context(hx, attr, &cx);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't create context");
		(*error_handler)(msg, st);
		free(msg);

		iiimcf_destroy_attr(attr);
		iiimcf_destroy_handle(hx);
		iiimcf_finalize();
		return NULL;
	}

	st = iiimcf_destroy_attr(attr);
	if(st != IIIMF_STATUS_SUCCESS) {
		msg = strdup("Can't destroy attribute");
		(*error_handler)(msg, st);
		free(msg);

		iiimcf_destroy_context(cx);
		iiimcf_destroy_handle(hx);
		iiimcf_finalize();
		return NULL;
	}

	return cx;
}

void
tii_destroy_context(IIIMCF_context cx)
{
	iiimcf_destroy_context(cx);
}

void
tii_destroy_handle(IIIMCF_handle hx)
{
	iiimcf_destroy_handle(hx);
}

void
tii_finalize_iiimcf(void)
{
	iiimcf_finalize();
}

void
tii_debug_print_event(IIIMCF_event ev)
{
	IIIMF_status st;
	IIIMCF_event_type et;

	st = iiimcf_get_event_type(ev, &et);
	if(st != IIIMF_STATUS_SUCCESS) {
		fprintf(stderr, "[event] Can't get event type\r\n");
		return;
	}

	switch(et) {
		case IIIMCF_EVENT_TYPE_DESTROY:
			fprintf(stderr, "[event] Destroy event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_RESET:
			fprintf(stderr, "[event] Reset event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_KEYEVENT:
			fprintf(stderr, "[event] Key event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_KEYEVENT_END:
			fprintf(stderr, "[event] Key event end\r\n");
			break;
		case IIIMCF_EVENT_TYPE_TRIGGER_NOTIFY:
			fprintf(stderr, "[event] Trigger notify event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_TRIGGER_NOTIFY_END:
			fprintf(stderr, "[event] Trigger notify event end\r\n");
			break;
		case IIIMCF_EVENT_TYPE_OPERATION:
			fprintf(stderr, "[event] Operation event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_SETICFOCUS:
			fprintf(stderr, "[event] Set IC focus event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UNSETICFOCUS:
			fprintf(stderr, "[event] Unset IC focus event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_EVENTLIKE_END:
			fprintf(stderr, "[event] Eventlike end\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI:
			fprintf(stderr, "[event] UI event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_PREEDIT:
			fprintf(stderr, "[event] UI preedit event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_PREEDIT_START:
			fprintf(stderr, "[event] UI preedit event start\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_PREEDIT_CHANGE:
			fprintf(stderr, "[event] UI preedit event change\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_PREEDIT_DONE:
			fprintf(stderr, "[event] UI preedit event done\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE:
			fprintf(stderr, "[event] UI lookup choice event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_START:
			fprintf(stderr, "[event] UI lookup choice event start\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_CHANGE:
			fprintf(stderr, "[event] UI lookup choice event change\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_LOOKUP_CHOICE_DONE:
			fprintf(stderr, "[event] UI lookup choice event done\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_STATUS_START:
			fprintf(stderr, "[event] UI status event start\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_STATUS_CHANGE:
			fprintf(stderr, "[event] UI status event change \r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_STATUS_DONE:
			fprintf(stderr, "[event] UI status event done\r\n");
			break;
		case IIIMCF_EVENT_TYPE_UI_COMMIT:
			fprintf(stderr, "[event] UI commit event\r\n");
			break;
		case IIIMCF_EVENT_TYPE_AUX_START:
			fprintf(stderr, "[event] AUX event start\r\n");
			break;
		case IIIMCF_EVENT_TYPE_AUX_DRAW:
			fprintf(stderr, "[event] AUX event draw\r\n");
			break;
		case IIIMCF_EVENT_TYPE_AUX_SETVALUES:
			fprintf(stderr, "[event] AUX event set values\r\n");
			break;
		case IIIMCF_EVENT_TYPE_AUX_DONE:
			fprintf(stderr, "[event] AUX event done\r\n");
			break;
		default:
			fprintf(stderr, "[event] Unknown event\r\n");
			break;
	}

	return;
}
