/*
 *  Terminal IIIMCF Interface/Terminal IIIMCF Input Method
 *  Copyright (c) 2005 Young Chol Song <nskystars at yahoo.com>
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer
 *     in the documentation and/or other materials provided with the
 *     distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 *  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 *  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*
 * Define input method and language
 * (refrence test program for more input methods)
 * change the following constants as shown to use other input methods
 *
 * Korean(hangul)  : HANGUL, IM "hangul",  LANG "ko"
 * Japanese(canna) : CANNA,  IM "CannaLE", LANG "ja"
 * Chinese(newpy)  : NEWPY,  IM "newpy",   LANG "zh_CN"
 */

#define HANGUL

#ifdef HANGUL 
#define IM "hangul"
#define LANG "ko"
#endif

#ifdef CANNA
#define IM "CannaLE"
#define LANG "ja"
#endif

#ifdef NEWPY
#define IM "newpy"
#define LANG "zh_CN"
#endif

#define _XOPEN_SOURCE

#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <poll.h>
#include <sys/ioctl.h>

/* tii/iiimf */
#include <iiimcf.h>
#include "tii.h"

/* terminfo */
#include <curses.h>
#include <term.h>


static volatile int keep = 1;
static volatile int winch_flag = 0;

static struct termios term_attr;

volatile int scr_height = 1;
volatile int scr_width = 1;
volatile int prev_scr_height = 1;

static int tty;
static int master;
static FILE * master_file;
static FILE * tty_file;

IIIMCF_handle handle;
IIIMCF_context context;

/* signals */
static void
chld_handler(int n)
{
	keep = 0;
}

static void
winch_handler(int n)
{
	struct winsize ws;

	prev_scr_height = scr_height;

	ioctl(tty, TIOCGWINSZ, &ws);
	scr_height = ws.ws_row;
	scr_width = ws.ws_col;
	if(ws.ws_row > 1)
		ws.ws_row--;
	ioctl(master, TIOCSWINSZ, &ws);

	winch_flag = 1;
}

/* terminfo putc */
int
tiim_ti_putc(int ch)
{
	fprintf(tty_file, "%c", ch);
	return ch;
}

/* user defined handler functions for tiim */
void
tiim_error_handler(char * str, int no)
{
	tputs(tigetstr("sc"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("sgr"), 0, 0, 0, 0, 0, 0, 0, 0, 0), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("cup"), scr_height - 1, 9), 1, tiim_ti_putc);
	tputs(tigetstr("el"), 1, tiim_ti_putc);
	fprintf(tty_file, "%s (%d)", str, no);
	tputs(tigetstr("rc"), 1, tiim_ti_putc);
	fflush(tty_file);
}

void
tiim_message_handler(char * str, char * item)
{
	tputs(tigetstr("sc"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("sgr"), 0, 0, 0, 0, 0, 0, 0, 0, 0), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("cup"), scr_height - 1, 9), 1, tiim_ti_putc);
	tputs(tigetstr("el"), 1, tiim_ti_putc);
	fprintf(tty_file, "%s (%s)", str, item);
	tputs(tigetstr("rc"), 1, tiim_ti_putc);
	fflush(tty_file);
}

void
tiim_lookup_handler(char * title_str, int win_choices, int win_rows, int win_cols,
			int win_dir, int ncand, int curr_idx, char ** cand_label, char ** cand_str)
{
	int i = 0;

	tputs(tigetstr("sc"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("sgr"), 0, 0, 0, 0, 0, 0, 0, 0, 0), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("cup"), scr_height - 1, 9), 1, tiim_ti_putc);
	tputs(tigetstr("el"), 1, tiim_ti_putc);
	fprintf(tty_file, "%s", title_str);
	for(i = 0; i < ncand; i++)
		if(i == curr_idx) {
			fprintf(tty_file, " ");
			tputs(tigetstr("smso"), 1, tiim_ti_putc);
			fprintf(tty_file, "[%s] %s", cand_label[i], cand_str[i]);
			tputs(tigetstr("rmso"), 1, tiim_ti_putc);
		}
		else
			fprintf(tty_file, " [%s] %s", cand_label[i], cand_str[i]);
	tputs(tigetstr("rc"), 1, tiim_ti_putc);
	fflush(tty_file);
}

void
tiim_preedit_handler(char * str, int cur_pos)
{
	tputs(tigetstr("sc"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("sgr"), 0, 0, 0, 0, 0, 0, 0, 0, 0), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("cup"), scr_height - 1, 9), 1, tiim_ti_putc);
	tputs(tigetstr("el"), 1, tiim_ti_putc);
	fprintf(tty_file, "%s", str);
	tputs(tigetstr("rc"), 1, tiim_ti_putc);
	fflush(tty_file);
}

void
tiim_commit_handler(char * str)
{
	fprintf(master_file, "%s", str);
	fflush(master_file);
}

void
tiim_status_handler(char * str)
{
	tputs(tigetstr("sc"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("sgr"), 0, 0, 0, 0, 0, 0, 0, 0, 0), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("cup"), scr_height - 1, 8), 1, tiim_ti_putc);
	tputs(tigetstr("el1"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("cup"), scr_height - 1, 0), 1, tiim_ti_putc);
	fprintf(tty_file, "%s", str);
	tputs(tigetstr("rc"), 1, tiim_ti_putc);
	fflush(tty_file);
}

void
tiim_keyevent_handler(int keycode, char keychar, int mod)
{
	fprintf(master_file, "%c", keychar);
	fflush(master_file);
}


void
tiim_atexit_uninit_ti(void)
{
	struct winsize ws;

	tcsetattr(tty, TCSADRAIN, &term_attr);

	ioctl(tty, TIOCGWINSZ, &ws);
	ioctl(master, TIOCSWINSZ, &ws);

	tputs(tigetstr("clear"), 1, tiim_ti_putc);
	tputs(tparm(tigetstr("csr"), 0, scr_height - 1), 1, tiim_ti_putc);
}

void
tiim_atexit_destroy_context(void)
{
	tii_destroy_context(context);
}

void
tiim_atexit_destroy_handle(void)
{
	tii_destroy_handle(handle);
}

void
tiim_atexit_uninit_iiimcf(void)
{
	tii_finalize_iiimcf();
}


int
main(void)
{
	struct pollfd fds[2];
	char * termtype;

	{
		struct sigaction act;

		act.sa_handler = chld_handler;
		act.sa_flags = 0;
		sigemptyset(&act.sa_mask);
		sigaction(SIGCHLD, &act, NULL);

		act.sa_handler = winch_handler;
		act.sa_flags = 0;
		sigemptyset(&act.sa_mask);
		sigaction(SIGWINCH, &act, NULL);
	}

	{
		tty = open("/dev/tty", O_RDWR);
		if(tty < 0) {
			perror("tty");
			exit(1);
		}

		master = open("/dev/ptmx", O_RDWR);
		if(master < 0) {
			perror("ptmx");
			exit(1);
		}

		if(grantpt(master)) {
			perror("grantpt");
			exit(1);
		}

		if(unlockpt(master)) {
			perror("unlockpt");
			exit(1);
		}

		if(fork() == 0) {
			char * name;
			int fd;
			char * shell, * shell_base;

			close(STDIN_FILENO);
			close(STDOUT_FILENO);
			close(STDERR_FILENO);

			setsid();
			name = ptsname(master);
			fd = open(name, O_RDWR);

			close(master);
			dup2(fd, STDIN_FILENO);
			dup2(fd, STDOUT_FILENO);
			dup2(fd, STDERR_FILENO);

			if(fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO)
				close(fd);
			shell = getenv("SHELL");
			if(shell == NULL)
				shell = "/bin/sh";
			shell_base = strrchr(shell, '/');
			if(shell_base == NULL)
				shell_base = shell;
			else
				shell_base++;

			execl(shell, shell_base, NULL);
			exit(1);
		}
	}

	termtype = getenv("TERM");
	setterm(termtype);

	{
		struct termios attr;
		tcgetattr(tty, &attr);
		term_attr = attr;
		attr.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
		attr.c_oflag &= ~OPOST;
		attr.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
		attr.c_cflag &= ~(CSIZE | PARENB);
		attr.c_cflag |= CS8;
		attr.c_cc[VMIN] = 1;
		tcsetattr(tty, TCSADRAIN, &attr);

		winch_handler(0);
		winch_flag = 0;

		master_file = fdopen(master, "wb");
		tty_file = fdopen(tty, "wb");
		fds[0].fd = tty;
		fds[1].fd = master;
		fds[0].events = fds[1].events = POLLIN | POLLHUP;

		tputs(tigetstr("clear"), 1, tiim_ti_putc);
		tputs(tparm(tigetstr("csr"), 0, scr_height - 2), 1, tiim_ti_putc);
		fflush(tty_file);

		atexit(tiim_atexit_uninit_ti);
	}

	{
		IIIMCF_input_method im;
		IIIMCF_language lang;

		tii_register_error_handler(tiim_error_handler);
		tii_register_message_handler(tiim_message_handler);
		tii_register_lookup_handler(tiim_lookup_handler);
		tii_register_preedit_handler(tiim_preedit_handler);
		tii_register_commit_handler(tiim_commit_handler);
		tii_register_status_handler(tiim_status_handler);
		tii_register_keyevent_handler(tiim_keyevent_handler);

		tii_init_iiimcf();
		atexit(tiim_atexit_uninit_iiimcf);

		handle = tii_create_handle();
		if(handle == NULL) {
			exit(-1);
		}
		atexit(tiim_atexit_destroy_handle);

		im = tii_find_im(handle, IM);
		lang = tii_find_lang(handle, IM, LANG);
		if(im == NULL || lang == NULL) {
			exit(-1);
		}

		context = tii_create_context(handle, im, lang, -1);
		if(context == NULL) {
			exit(-1);
		}
		atexit(tiim_atexit_destroy_context);

		tii_fwd_trigger(context, 1);
		tii_fwd_trigger(context, 0);

		tiim_message_handler("IIIMCF Initialization success", "Welcome to Terminal IIIMCF Input Method");
	}

	while(keep) {
		if(poll(fds, 2, -1) <= 0)
			continue;
		if(winch_flag) {
			tputs(tparm(tigetstr("csr"), 0, scr_height - 2), 1, tiim_ti_putc);
			tputs(tparm(tigetstr("cup"), scr_height - 1, 0), 1, tiim_ti_putc);
			tputs(tigetstr("el"), 1, tiim_ti_putc);
			tputs(tparm(tigetstr("cup"), scr_height - 2, 0), 1, tiim_ti_putc);
			if(prev_scr_height > scr_height)
				tputs(tigetstr("cud1"), 1, tiim_ti_putc);
			fflush(tty_file);

			winch_flag = 0;
		}
		if((fds[0].revents & POLLHUP) || (fds[1].revents & POLLHUP))
			break;
		if(fds[0].revents & POLLIN) {
			unsigned char b[4096];
			int l, i, shift_state;
			l = read(tty, b, sizeof(b));
			if(l < 0)
				exit(1);
			for(i = 0; keep && i < l; i++) {
				if(b[i] == 0) {
					if(tii_get_conversion_mode(context) == 1) {
						tii_fwd_trigger(context, 0);
						tiim_status_handler("");
					}
					else
						tii_fwd_trigger(context, 1);
					tii_process_events(context);
					continue;
				}
				shift_state = 0;
				if(b[i] <= 90 && b[i] >= 65)
					shift_state = 1;
				if(!tii_fwd_key(context, 0, b[i], shift_state)) {
					putc(b[i], master_file);
					continue;
				}
				if(!tii_process_events(context)) exit(-1); 
				//putc(b[i], master_file);
			}
			fflush(master_file);
		}
		if(fds[1].revents & POLLIN) {
			unsigned char b[4096];
			int l;
			l = read(master, b, sizeof(b));
			if(l < 0)
				exit(1);
			if(l > 0)
				write(tty, b, l);
		}
	}

	return 0;
}
